unit Unit1;

interface

{$IFDEF VER150}
{$WARN UNSAFE_CAST OFF}
{$WARN UNSAFE_TYPE OFF}
{$ENDIF}
{$IFDEF VER170}
{$WARN UNSAFE_CAST OFF}
{$WARN UNSAFE_CODE OFF}
{$ENDIF}

uses
{$IFDEF VER140}
  Variants,
{$ENDIF}
{$IFDEF VER150}
  Variants,
{$ENDIF}
{$IFDEF VER170}
  Variants,
{$ENDIF}
{$IFDEF VER180}
  Variants,
{$ENDIF}
  Windows, Forms, ComObj, ActiveX, SysUtils, OPCtypes, OPCDA, OPCutils, Classes,
  Dialogs, Controls, StdCtrls, ExtCtrls, Grids,

  OPCenum;
type
 // class to receive IDataObject data change advises
  TOPCAdviseSink = class(TInterfacedObject, IAdviseSink)
  public
    procedure OnDataChange(const formatetc: TFormatEtc;
                           const stgmed: TStgMedium); stdcall;
    procedure OnViewChange(dwAspect: Longint; lindex: Longint); stdcall;
    procedure OnRename(const mk: IMoniker); stdcall;
    procedure OnSave; stdcall;
    procedure OnClose; stdcall;
  end;

  // class to receive IConnectionPointContainer data change callbacks
  TOPCDataCallback = class(TInterfacedObject, IOPCDataCallback)
  public
    function OnDataChange(dwTransid: DWORD; hGroup: OPCHANDLE;
      hrMasterquality: HResult; hrMastererror: HResult; dwCount: DWORD;
      phClientItems: POPCHANDLEARRAY; pvValues: POleVariantArray;
      pwQualities: PWordArray; pftTimeStamps: PFileTimeArray;
      pErrors: PResultList): HResult; stdcall;
    function OnReadComplete(dwTransid: DWORD; hGroup: OPCHANDLE;
      hrMasterquality: HResult; hrMastererror: HResult; dwCount: DWORD;
      phClientItems: POPCHANDLEARRAY; pvValues: POleVariantArray;
      pwQualities: PWordArray; pftTimeStamps: PFileTimeArray;
      pErrors: PResultList): HResult; stdcall;
    function OnWriteComplete(dwTransid: DWORD; hGroup: OPCHANDLE;
      hrMastererr: HResult; dwCount: DWORD; pClienthandles: POPCHANDLEARRAY;
      pErrors: PResultList): HResult; stdcall;
    function OnCancelComplete(dwTransid: DWORD; hGroup: OPCHANDLE):
      HResult; stdcall;
  end;

  TForm1 = class(TForm)
    Memo1: TMemo;
    Label1: TLabel;
    Label2: TLabel;
    Edit1: TEdit;
    Edit2: TEdit;
    Timer1: TTimer;
    Button1: TButton;
    StringGrid1: TStringGrid;
    Button2: TButton;
    Edit3: TEdit;
    Label3: TLabel;
    Button3: TButton;
    procedure FormCreate(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
  private
    { Private-Deklarationen }
    InTimer1: Boolean;
    TimeOut : Boolean;
    procedure CleanUp(Sender: TObject);
  public
    { Public-Deklarationen }
  end;

const
  ServerProgID = 'FirstServ.DA2';  //MRD.DA2  //'FirstServ.DA2'  //'MRD.DA2.1'
  Host         = 'palmsamsung';

  OneSecond = 1 / (24 * 60 * 60);

  RPC_C_AUTHN_LEVEL_NONE = 1;
  RPC_C_IMP_LEVEL_IMPERSONATE = 3;
  EOAC_NONE = 0;

var
  Form1          : TForm1;
  ServerIf       : IOPCServer;
  GroupIf        : IOPCItemMgt;
  GroupHandle    : OPCHANDLE;
  ItemHandle     : Array [1..14] of OPCHANDLE;
  ItemType       : TVarType;
  ItemValue      : string;
  ItemQuality    : Word;
  AdviseSink     : IAdviseSink;
  AsyncConnection: Longint;
  OPCDataCallback: IOPCDataCallback;
  StartTime      : TDateTime;

  OPCServerList  : TOPCServerList;
  CATIDs         : array of TGUID;
  ServerNames    : TStringList;

implementation

{$R *.dfm}

// TOPCAdviseSink methods

// OPC standard says this is the only method we need to fill in
procedure TOPCAdviseSink.OnDataChange(const formatetc: TFormatEtc;
                                      const stgmed   : TStgMedium);
var
  PG          : POPCGROUPHEADER;
  PI1         : POPCITEMHEADER1ARRAY;
  PI2         : POPCITEMHEADER2ARRAY;
  PV          : POleVariant;
  I           : Integer;
  PStr        : PWideChar;
  NewValue    : string;
  WithTime    : Boolean;
  ClientHandle: OPCHANDLE;
  Quality     : Word;

begin
  // the rest of this method assumes that the item header array uses
  // OPCITEMHEADER1 or OPCITEMHEADER2 records,
  // so check this first to be defensive
  if (formatetc.cfFormat <> OPCSTMFORMATDATA) and
      (formatetc.cfFormat <> OPCSTMFORMATDATATIME) then Exit;
  // does the data stream provide timestamps with each value?
  WithTime := formatetc.cfFormat = OPCSTMFORMATDATATIME;

  PG := GlobalLock(stgmed.hGlobal);
  if PG <> nil then
  begin
    // we will only use one of these two values, according to whether
    // WithTime is set:
    PI1 := Pointer(PChar(PG) + SizeOf(OPCGROUPHEADER));
    PI2 := Pointer(PI1);
    if Succeeded(PG.hrStatus) then
    begin
      for I := 0 to PG.dwItemCount - 1 do
      begin
        if WithTime then
        begin
          PV := POleVariant(PChar(PG) + PI1[I].dwValueOffset);
          ClientHandle := PI1[I].hClient;
          Quality := (PI1[I].wQuality and OPC_QUALITY_MASK);
        end
        else begin
          PV := POleVariant(PChar(PG) + PI2[I].dwValueOffset);
          ClientHandle := PI2[I].hClient;
          Quality := (PI2[I].wQuality and OPC_QUALITY_MASK);
        end;
        if Quality = OPC_QUALITY_GOOD then
        begin
          // this test assumes we're not dealing with array data
          if TVarData(PV^).VType <> VT_BSTR then
          begin
            NewValue := VarToStr(PV^);
          end
          else begin
            // for BSTR data, the BSTR image follows immediately in the data
            // stream after the variant union;  the BSTR begins with a DWORD
            // character count, which we skip over as the BSTR is also
            // NULL-terminated
            PStr := PWideChar(PChar(PV) + SizeOf(OleVariant) + 4);
            NewValue := WideString(PStr);
          end;
          if WithTime then
          begin
            Writeln('New value for item ', ClientHandle, ' advised:  ',
                    NewValue, ' (with timestamp)');
          end
          else begin
            Writeln('New value for item ', ClientHandle, ' advised:  ',
                    NewValue);
          end;
        end
        else begin
          Writeln('Advise received for item ', ClientHandle,
                  ' , but quality not good');
        end;
      end;
    end;
    GlobalUnlock(stgmed.hGlobal);
  end;
end;

procedure TOPCAdviseSink.OnViewChange(dwAspect: Longint; lindex: Longint);
begin
end;

procedure TOPCAdviseSink.OnRename(const mk: IMoniker);
begin
end;

procedure TOPCAdviseSink.OnSave;
begin
end;

procedure TOPCAdviseSink.OnClose;
begin
end;

// TOPCDataCallback methods

function TOPCDataCallback.OnDataChange(dwTransid: DWORD; hGroup: OPCHANDLE;
  hrMasterquality: HResult; hrMastererror: HResult; dwCount: DWORD;
  phClientItems: POPCHANDLEARRAY; pvValues: POleVariantArray;
  pwQualities: PWordArray; pftTimeStamps: PFileTimeArray;
  pErrors: PResultList): HResult;
var
  ClientItems: POPCHANDLEARRAY;
  Values: POleVariantArray;
  Qualities: PWORDARRAY;
  I: Integer;
  NewValue: string;
begin
  Result := S_OK;
  ClientItems := POPCHANDLEARRAY(phClientItems);
  Values := POleVariantArray(pvValues);
  Qualities := PWORDARRAY(pwQualities);
  for I := 0 to dwCount - 1 do
  begin
    if Qualities[I] = OPC_QUALITY_GOOD then
    begin
      NewValue := VarToStr(Values[I]);
      Writeln('New callback for item ', ClientItems[I], ' received, value:  ',
              NewValue);
    end
    else begin
      Writeln('Callback received for item ', ClientItems[I],
              ' , but quality not good');
    end;
  end;
end;

function TOPCDataCallback.OnReadComplete(dwTransid: DWORD; hGroup: OPCHANDLE;
  hrMasterquality: HResult; hrMastererror: HResult; dwCount: DWORD;
  phClientItems: POPCHANDLEARRAY; pvValues: POleVariantArray;
  pwQualities: PWordArray; pftTimeStamps: PFileTimeArray;
  pErrors: PResultList): HResult;
begin
  Result := OnDataChange(dwTransid, hGroup, hrMasterquality, hrMastererror,
    dwCount, phClientItems, pvValues, pwQualities, pftTimeStamps, pErrors);
end;

function TOPCDataCallback.OnWriteComplete(dwTransid: DWORD; hGroup: OPCHANDLE;
  hrMastererr: HResult; dwCount: DWORD; pClienthandles: POPCHANDLEARRAY;
  pErrors: PResultList): HResult;
begin
  // we don't use this facility
  Result := S_OK;
end;

function TOPCDataCallback.OnCancelComplete(dwTransid: DWORD;
                                           hGroup   : OPCHANDLE): HResult;
begin
 // we don't use this facility
 Result := S_OK;
end;

Procedure TForm1.CleanUp(Sender: TObject);
var
  HR: HResult;
begin
 //now cleanup
 try
  HR:= ServerIf.RemoveGroup(GroupHandle, False);
  if Succeeded(HR)
   then Memo1.Lines.Add('Removed group')
   else Memo1.Lines.Add('Unable to remove group');
 except
  //in case OPC is not yet connected
 end;
 Edit3.Text:= 'Disconnected';
end;

procedure TForm1.Button1Click(Sender: TObject);
var
 StartTime: TDateTime;
begin
 Timer1.Enabled:= False;
 Memo1.Lines.Add('Waiting 2 seconds for Timer to quit');
 StartTime := Now;
 while ((Now - StartTime) < (2 * OneSecond)) or (InTimer1) do Application.ProcessMessages;

 CleanUp(Self);
 Application.Terminate;
end;

Function IsInIDE: Boolean;
begin
 Result:= (FindWindow('TAppBuilder', nil) > 0);
end;

procedure TForm1.Button2Click(Sender: TObject);
var
 HR: HResult;
 i : Integer;
Label
 Exi;
begin
 // this is for DCOM:
 // without this, callbacks from the server may get blocked, depending on
 // DCOM configuration settings
 If IsInIDE then
  //skip
 else
  begin
   HR:= CoInitializeSecurity(
    nil,                    // points to security descriptor
    -1,                     // count of entries in asAuthSvc
    nil,                    // array of names to register
    nil,                    // reserved for future use
    RPC_C_AUTHN_LEVEL_NONE, // the default authentication level for proxies
    RPC_C_IMP_LEVEL_IMPERSONATE,// the default impersonation level for proxies
    nil,                    // used only on Windows 2000
    EOAC_NONE,              // additional client or server-side capabilities
    nil                     // reserved for future use
    );
  if Failed(HR) then
  begin
   Memo1.Lines.Add('Failed to initialize DCOM security');
   Goto Exi;
  end;
 end;
 Edit3.Text:= 'Try to Connect';
 If Host > ''
  then
   Memo1.Lines.Add('Try to connect to \\' + Edit2.Text + '\' + Edit1.Text)
  else
   Memo1.Lines.Add('Try to connect to ' + Edit1.Text);
 try
   //we will use the custom OPC interfaces, and OPCProxy.dll will handle
   //marshaling for us automatically (if registered)
   ServerIf:= CreateRemoteComObject(Edit2.Text, ProgIDToClassID(Edit1.Text)) as IOPCServer;
 except
   ServerIf:= nil;
 end;
  if ServerIf <> nil
   then
    begin
     Memo1.Lines.Add('Connected to OPC server');
     Edit3.Text:= 'Connected';
    end
   else
    begin
     Memo1.Lines.Add('Unable to connect to OPC server');
     Goto Exi;
    end;

  //now add 2 groups

  HR:= ServerAddGroup(ServerIf, 'MyGroup', True, 700, 0, GroupIf, GroupHandle);
  if Succeeded(HR)
   then
    Memo1.Lines.Add('Added group to server')
   else
    begin
     Memo1.Lines.Add('Unable to add group to server');
     Goto Exi;
    end;

  // now add an items to group #1
  For i:= 1 to 14 do
  begin
   HR:= GroupAddItem(GroupIf, StringGrid1.Cells[1, i], 0, VT_EMPTY,
                     ItemHandle[i], ItemType);
   if Succeeded(HR)
    then
     Memo1.Lines.Add('Added item #' + IntToStr(i) +' to group')
    else
     begin
      Memo1.Lines.Add('Unable to add item #' + IntToStr(i) +
                      ' "' + StringGrid1.Cells[1, i] + '" ' +'to group');
      ServerIf.RemoveGroup(GroupHandle, False);
      Goto Exi;
     end;
  end;

 {
  // set up an IDataObject advise callback for the group
  AdviseSink:= TOPCAdviseSink.Create;
  HR:= GroupAdviseTime(GroupIf, AdviseSink, AsyncConnection);
  if Failed(HR)
   then
    Memo1.Lines.Add('Failed to set up IDataObject advise callback')
   else
    begin
     Memo1.Lines.Add('IDataObject advise callback established');
     // continue waiting for callbacks for 10 seconds
     StartTime:= Now;
     While (Now - StartTime) < (10 * OneSecond) do Application.ProcessMessages;
     // end the IDataObject advise callback
     GroupUnadvise(GroupIf, AsyncConnection);
    end;

  // now set up an IConnectionPointContainer data callback for the group
  OPCDataCallback:= TOPCDataCallback.Create;
  HR:= GroupAdvise2(GroupIf, OPCDataCallback, AsyncConnection);
  if Failed(HR)
   then
    Memo1.Lines.Add('Failed to set up IConnectionPointContainer advise callback')
   else
    begin
     Memo1.Lines.Add('IConnectionPointContainer data callback established');
     // continue waiting for callbacks for 10 seconds
     StartTime:= Now;
     While (Now - StartTime) < (10 * OneSecond) do Application.ProcessMessages;
     // end the IConnectionPointContainer data callback
     GroupUnadvise2(GroupIf, AsyncConnection);
    end;
  }
 Timer1.Enabled:= True;

Exi:
 {
 ShowMessage('Will terminate program');
 Button1Click(self);
 }
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
 InTimer1:= False;
 TimeOut:= False;

 // among other things, this call makes sure that COM is initialized
 Memo1.Clear;
 Edit1.Text:= ServerProgID;

 Stringgrid1.Cells[0, 0]:= 'Prop ID';
 Stringgrid1.Cells[1, 0]:= 'Tag Name';
 Stringgrid1.Cells[2, 0]:= 'Data Type';
 Stringgrid1.Cells[3, 0]:= 'Actual Value';

 Stringgrid1.Cells[0, 1]:= '5000';
 Stringgrid1.Cells[1, 1]:= 'Complete';
 Stringgrid1.Cells[2, 1]:= 'VT_BSTR';

 Stringgrid1.Cells[0, 2]:= '5001';
 Stringgrid1.Cells[1, 2]:= 'Date.Complete';
 Stringgrid1.Cells[2, 2]:= 'VT_BSTR';

 Stringgrid1.Cells[0, 3]:= '5002';
 Stringgrid1.Cells[1, 3]:= 'Date.Parts.Day';
 Stringgrid1.Cells[2, 3]:= 'VT_UI2';

 Stringgrid1.Cells[0, 4]:= '5003';
 Stringgrid1.Cells[1, 4]:= 'Date.Parts.Month';
 Stringgrid1.Cells[2, 4]:= 'VT_UI2';

 Stringgrid1.Cells[0, 5]:= '5004';
 Stringgrid1.Cells[1, 5]:= 'Date.Parts.Year';
 Stringgrid1.Cells[2, 5]:= 'VT_UI2';

 Stringgrid1.Cells[0, 6]:= '5005';
 Stringgrid1.Cells[1, 6]:= 'Time.Complete';
 Stringgrid1.Cells[2, 6]:= 'VT_BSTR';

 Stringgrid1.Cells[0, 7]:= '5006';
 Stringgrid1.Cells[1, 7]:= 'Time.Parts.Hour';
 Stringgrid1.Cells[2, 7]:= 'VT_UI2';

 Stringgrid1.Cells[0, 8]:= '5007';
 Stringgrid1.Cells[1, 8]:= 'Time.Parts.Minute';
 Stringgrid1.Cells[2, 8]:= 'VT_UI2';

 Stringgrid1.Cells[0, 9]:= '5008';
 Stringgrid1.Cells[1, 9]:= 'Time.Parts.Second';
 Stringgrid1.Cells[2, 9]:= 'VT_UI2';

 Stringgrid1.Cells[0, 10]:= '5009';
 Stringgrid1.Cells[1, 10]:= 'Time.Parts.Millisecond';
 Stringgrid1.Cells[2, 10]:= 'VT_UI2';

 Stringgrid1.Cells[0, 11]:= '5010';
 Stringgrid1.Cells[1, 11]:= 'Test_Tag_1.Actual';
 Stringgrid1.Cells[2, 11]:= 'VT_UI2';

 Stringgrid1.Cells[0, 12]:= '5011';
 Stringgrid1.Cells[1, 12]:= 'Test_Tag_1.Inverted';
 Stringgrid1.Cells[2, 12]:= 'VT_UI2';

 Stringgrid1.Cells[0, 13]:= '5012';
 Stringgrid1.Cells[1, 13]:= 'Test_Tag_2.Actual';
 Stringgrid1.Cells[2, 13]:= 'VT_UI2';

 Stringgrid1.Cells[0, 14]:= '5013';
 Stringgrid1.Cells[1, 14]:= 'Test_Tag_2.Inverted';
 Stringgrid1.Cells[2, 14]:= 'VT_UI2';

 Self.Show;
 Edit2.Text:= Host;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
 HR: HResult;
 i : Integer;
begin
 InTimer1:= True;
 Timer1.Enabled:= False;
 // now try to read the item value synchronously
 For i:= 1 to 14 do
 begin
  HR:= ReadOPCGroupItemValue(GroupIf, ItemHandle[i], ItemValue, ItemQuality);
  if Succeeded(HR) then
  begin
   if (ItemQuality and OPC_QUALITY_MASK) = OPC_QUALITY_GOOD
    then
     StringGrid1.Cells[3, i]:= ItemValue
    else
     Memo1.Lines.Add('Item ' + IntToStr(i) +
                     ' value was read synchronously, but quality was poor');
   Timer1.Enabled:= True;
  end
  else
   begin
    Memo1.Lines.Add('Failed to read item ' + IntToStr(i) +
                    ' value synchronously');

    CleanUp(Self);
    Break;
   end;
 end;

 {
  // finally write the value just read from item 7 into item 3
  HR:= WriteOPCGroupItemValue(GroupIf, ItemHandle[3],
                              StrToInt(StringGrid1.Cells[3, 7]));
  if Succeeded(HR)
   then
    Memo1.Lines.Add('Item 3 value written synchronously')
   else
    Memo1.Lines.Add('Failed to write item 3 value synchronously');

  // wait for 1 second
  StartTime:= Now;
  while (Now - StartTime) < OneSecond do Application.ProcessMessages;

 // and try to read it back
 HR:= ReadOPCGroupItemValue(GroupIf, Item1Handle, ItemValue, ItemQuality);
 if Succeeded(HR) then
 begin
  if (ItemQuality and OPC_QUALITY_MASK) = OPC_QUALITY_GOOD
   then
    Edit6.Text:= ItemValue
   else
    Memo1.Lines.Add('Item 1 value was read synchronously, but quality was not good');
 end
 else
  Memo1.Lines.Add('Failed to read item 0 value synchronously');
 }
 InTimer1:= False;
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  i: Integer;
begin //Enumerate available Servers
 ServerNames:= TStringList.Create;
 try
  SetLength(CATIDs, 0);
  OPCServerList:= TOPCServerList.Create(Host, True, CATIDs);
  try
   Memo1.Lines.Add('Enumerating OPC Servers...May take several minutes!!');
   Memo1.Lines.Add('');
   OPCServerList.Update;
   ServerNames.AddStrings(OPCServerList.Items);
  finally
   OPCServerList.Free;
  end;
  Memo1.Lines.Add('OPC servers from registry: -');
  for I:= 0 to ServerNames.Count - 1 do Memo1.Lines.Add(ServerNames[I]);

  Memo1.Lines.Add('');
  ServerNames.Clear;

  SetLength(CATIDs, 1);
  CATIDs[0]:= CATID_OPCDAServer20;
  OPCServerList:= TOPCServerList.Create('', False, CATIDs);
  try
   OPCServerList.Update;
   ServerNames.AddStrings(OPCServerList.Items);
  finally
   OPCServerList.Free;
  end;
  Memo1.Lines.Add('OPC DA 2.0 servers from server browser: -');
  for I:= 0 to ServerNames.Count - 1 do Memo1.Lines.Add(ServerNames[I]);
  Memo1.Lines.Add('');
  finally
   ServerNames.Free;
  end;
end;

end.
